"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import datetime
import sys
from decimal import Decimal
from ipaddress import ip_address

import pytest
from tcolorpy import tcolor
from typepy import (
    Bool,
    DateTime,
    Dictionary,
    Infinity,
    Integer,
    IpAddress,
    List,
    Nan,
    NoneType,
    NullString,
    RealNumber,
    String,
    Typecode,
)

from dataproperty import Align, ColumnDataProperty, DataProperty, Format


nan = float("nan")
inf = float("inf")


class Test_ColumnDataPeroperty:
    DATATIME_DATA = datetime.datetime(2017, 1, 1, 1, 2, 3)

    @pytest.mark.parametrize(
        ["values", "expected_typecode", "expected_class"],
        [
            # single type values
            [[None, None], Typecode.NONE, NoneType],
            [[0, sys.maxsize, str(sys.maxsize), -sys.maxsize], Typecode.INTEGER, Integer],
            [[0, 1.1, "0.01", -sys.maxsize], Typecode.REAL_NUMBER, RealNumber],
            [["-0.538882625371217", "0.268624155343302", ""], Typecode.REAL_NUMBER, RealNumber],
            [[ip_address("127.0.0.1"), ip_address("::1")], Typecode.IP_ADDRESS, IpAddress],
            [[0, 1.1, -sys.maxsize, "test"], Typecode.STRING, String],
            [["", ""], Typecode.NULL_STRING, NullString],
            [[True, True, False], Typecode.BOOL, Bool],
            [[True, "True", False], Typecode.STRING, String],
            [[DATATIME_DATA, DATATIME_DATA], Typecode.DATETIME, DateTime],
            [["2017-01-01 00:00:00", "2017-01-02 03:04:05+09:00"], Typecode.STRING, String],
            [[inf, "inf", "infinity", "INF"], Typecode.INFINITY, Infinity],
            [[nan, "nan", "NAN"], Typecode.NAN, Nan],
            [[{"a": 1}, {"b": 2}], Typecode.DICTIONARY, Dictionary],
            [[[1, 2], ["a", "b"]], Typecode.LIST, List],
            # not mixed types with None value
            [[None, sys.maxsize, str(-sys.maxsize)], Typecode.INTEGER, Integer],
            [[1, None, ""], Typecode.INTEGER, Integer],
            [[1.1, None], Typecode.REAL_NUMBER, RealNumber],
            [[1.1, None, ""], Typecode.REAL_NUMBER, RealNumber],
            [[0, 1.1, Decimal("0.1"), None, ""], Typecode.REAL_NUMBER, RealNumber],
            [
                [ip_address("192.168.0.1"), None, ip_address("::1"), None],
                Typecode.IP_ADDRESS,
                IpAddress,
            ],
            [[None, "test"], Typecode.STRING, String],
            [[None, True, False], Typecode.BOOL, Bool],
            [[None, True, "False"], Typecode.STRING, String],
            [[None, DATATIME_DATA, None], Typecode.DATETIME, DateTime],
            [[None, inf], Typecode.INFINITY, Infinity],
            [[None, nan], Typecode.NAN, Nan],
            # mixed types
            [[True, 1], Typecode.STRING, String],
            [[DATATIME_DATA, "test"], Typecode.STRING, String],
            [[inf, 0.1], Typecode.REAL_NUMBER, RealNumber],
            [[inf, "test"], Typecode.STRING, String],
            [[nan, 0.1], Typecode.REAL_NUMBER, RealNumber],
            [[nan, "test"], Typecode.STRING, String],
            [[sys.maxsize, inf, nan], Typecode.REAL_NUMBER, RealNumber],
            [[DATATIME_DATA, str(DATATIME_DATA), DATATIME_DATA], Typecode.STRING, String],
            [[1, 1.1, DATATIME_DATA, "test", None, True, inf, Nan], Typecode.STRING, String],
        ],
    )
    def test_normal_typecode_type_class(self, values, expected_typecode, expected_class):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("dummy"))

        for value in values:
            col_dp.update_body(DataProperty(value))

        assert col_dp.typecode == expected_typecode
        assert col_dp.type_class == expected_class

    def test_normal_number_0(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [0, -1.234, 55.55]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 3
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 6

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 2

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 3

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 1

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=6, "
            "int_digits=(min=1, max=2), decimal_places=(min=0, max=3), "
            "extra_len=(min=0, max=1)"
        )

    def test_normal_number_1(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [0, inf, nan]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 0
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 8

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 0

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=8, "
            "int_digits=1, decimal_places=0"
        )

    def test_normal_number_2(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [1, 2.2, -3]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 1
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 4

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 1

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 1

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=4, "
            "int_digits=1, decimal_places=(min=0, max=1), "
            "extra_len=(min=0, max=1)"
        )

    def test_normal_number_3(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [0.01, 2.2, None]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 2
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 4

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 1
        assert col_dp.minmax_decimal_places.max_value == 2

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=4, "
            "int_digits=1, decimal_places=(min=1, max=2)"
        )

    def test_normal_number_4(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [0.01, 1.0, 1.2]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 2
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 4

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 2
        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=4, "
            "int_digits=1, decimal_places=(min=0, max=2)"
        )

    def test_normal_number_5(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [1.1, 2.2, 3.33]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 2
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 4

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 1
        assert col_dp.minmax_decimal_places.max_value == 2
        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=4, "
            "int_digits=1, decimal_places=(min=1, max=2)"
        )

    def test_normal_text_number(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in ["1,000,000,000,000", "1"]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 0
        assert col_dp.typecode == Typecode.INTEGER
        assert col_dp.ascii_char_width == 13

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 13

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 0
        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == (
            "column=0, type=INTEGER, align=right, ascii_width=13, "
            "bit_len=40, int_digits=(min=1, max=13), decimal_places=0"
        )

    def test_normal_inf(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("inf"))

        for value in [inf, None, inf, "inf"]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.LEFT
        assert col_dp.decimal_places is None
        assert col_dp.typecode == Typecode.INFINITY
        assert col_dp.ascii_char_width == 8

        assert col_dp.minmax_integer_digits.min_value is None
        assert col_dp.minmax_integer_digits.max_value is None

        assert col_dp.minmax_decimal_places.min_value is None
        assert col_dp.minmax_decimal_places.max_value is None

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == ("column=0, type=INFINITY, align=left, ascii_width=8")

    def test_normal_mix_0(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [0, -1.234, 55.55, "abcdefg"]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.LEFT
        assert col_dp.decimal_places == 3
        assert col_dp.typecode == Typecode.STRING
        assert col_dp.ascii_char_width == 7

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 2

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 3

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 1

        assert str(col_dp) == (
            "column=0, type=STRING, align=left, ascii_width=7, "
            "int_digits=(min=1, max=2), decimal_places=(min=0, max=3), "
            "extra_len=(min=0, max=1)"
        )

    def test_normal_number_ansi_escape(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [
            tcolor("1", color="red"),
            tcolor("2.2", color="green"),
            tcolor("-3", color="blue"),
        ]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 1
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 4

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 1

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 1

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=4, "
            "int_digits=1, decimal_places=(min=0, max=1), "
            "extra_len=(min=0, max=1)"
        )

    def test_normal_mix_ansi_escape(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [
            tcolor("0", color="red"),
            tcolor("-1.234", color="yellow"),
            tcolor("55.55", color="green"),
            tcolor("abcdefg", color="blue"),
        ]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.LEFT
        assert col_dp.decimal_places == 3
        assert col_dp.typecode == Typecode.STRING
        assert col_dp.ascii_char_width == 7

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 2

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 3

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 1

        assert str(col_dp) == (
            "column=0, type=STRING, align=left, ascii_width=7, "
            "int_digits=(min=1, max=2), decimal_places=(min=0, max=3), "
            "extra_len=(min=0, max=1)"
        )

    @pytest.mark.parametrize(
        ["values", "expected"],
        [
            [[0, 1, 0, 1], 1],
            [[-128, 0, 127, None], 8],
        ],
    )
    def test_normal_bit_length(self, values, expected):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("dummy"))

        for value in values:
            col_dp.update_body(DataProperty(value))

        assert col_dp.typecode == Typecode.INTEGER
        assert col_dp.bit_length == expected

    @pytest.mark.parametrize(
        ["values", "expected"],
        [
            [[0.1, 1], None],
            [["aaa", "0.0.0.0"], None],
        ],
    )
    def test_abnormal_bit_length(self, values, expected):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("dummy"))

        for value in values:
            col_dp.update_body(DataProperty(value))

        assert col_dp.bit_length == expected

    def test_normal_multibyte_char(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in ["いろは", "abcde"]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.LEFT
        assert col_dp.decimal_places is None
        assert col_dp.typecode == Typecode.STRING
        assert col_dp.ascii_char_width == 6

        assert col_dp.minmax_integer_digits.min_value is None
        assert col_dp.minmax_integer_digits.max_value is None

        assert col_dp.minmax_decimal_places.min_value is None
        assert col_dp.minmax_decimal_places.max_value is None

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

        assert str(col_dp) == ("column=0, type=STRING, align=left, ascii_width=6")

    @pytest.mark.parametrize(
        ["ambiguous_width", "ascii_char_width"],
        [
            [2, 6],
            [1, 3],
        ],
    )
    def test_normal_east_asian_ambiguous_width(self, ambiguous_width, ascii_char_width):
        col_dp = ColumnDataProperty(
            0, float_type=Decimal, east_asian_ambiguous_width=ambiguous_width
        )
        col_dp.update_header(DataProperty("abc"))

        for value in ["ØØØ", "α", "ββ"]:
            col_dp.update_body(DataProperty(value, east_asian_ambiguous_width=ambiguous_width))

        assert col_dp.align == Align.LEFT
        assert col_dp.decimal_places is None
        assert col_dp.typecode == Typecode.STRING
        assert col_dp.ascii_char_width == ascii_char_width

        assert col_dp.minmax_integer_digits.min_value is None
        assert col_dp.minmax_integer_digits.max_value is None

        assert col_dp.minmax_decimal_places.min_value is None
        assert col_dp.minmax_decimal_places.max_value is None

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

    def test_normal_max_precision(self):
        col_dp = ColumnDataProperty(column_index=0, float_type=Decimal, max_precision=3)
        col_dp.update_header(DataProperty("abc"))

        for value in ["0.0000000000001", "0.1", "0"]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 3
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == 5

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 1

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 13

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 0

    def test_min_width(self):
        min_width = 100

        col_dp = ColumnDataProperty(0, min_width=min_width, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        for value in [0, -1.234, 55.55]:
            col_dp.update_body(DataProperty(value))

        assert col_dp.align == Align.RIGHT
        assert col_dp.decimal_places == 3
        assert col_dp.typecode == Typecode.REAL_NUMBER
        assert col_dp.ascii_char_width == min_width

        assert col_dp.minmax_integer_digits.min_value == 1
        assert col_dp.minmax_integer_digits.max_value == 2

        assert col_dp.minmax_decimal_places.min_value == 0
        assert col_dp.minmax_decimal_places.max_value == 3

        assert col_dp.minmax_additional_format_len.min_value == 0
        assert col_dp.minmax_additional_format_len.max_value == 1

        assert str(col_dp) == (
            "column=0, type=REAL_NUMBER, align=right, ascii_width=100, "
            "int_digits=(min=1, max=2), decimal_places=(min=0, max=3), "
            "extra_len=(min=0, max=1)"
        )

    def test_extend_width(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        col_dp.update_header(DataProperty("abc"))

        assert col_dp.ascii_char_width == 3

        col_dp.extend_width(2)

        assert col_dp.ascii_char_width == 5

    def test_null(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        assert col_dp.align == Align.LEFT
        assert col_dp.decimal_places is None
        assert col_dp.typecode == Typecode.NONE
        assert col_dp.ascii_char_width == 0


class Test_ColumnDataPeroperty_dp_to_str:
    def test_normal_0(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        values = [0.1, 3.4375, 65.5397978633, 189.74439359, 10064.0097539, "abcd"]
        expected_list = [
            "0.1000000000",
            "3.4375000000",
            "65.5397978633",
            "189.7443935900",
            "10064.0097539000",
            "abcd",
        ]

        col_dp.update_header(DataProperty("abc"))
        for value in values:
            col_dp.update_body(DataProperty(value))

        assert len(values) == len(expected_list)
        for value, expected in zip(values, expected_list):
            assert col_dp.dp_to_str(DataProperty(value)) == expected

    def test_normal_1(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        values = [0, 0.1]
        expected_list = ["0", "0.1"]

        col_dp.update_header(DataProperty("abc"))
        for value in ["abcd", "efg"]:
            col_dp.update_body(DataProperty(value))

        assert len(values) == len(expected_list)
        for value, expected in zip(values, expected_list):
            assert col_dp.dp_to_str(DataProperty(value)) == expected

    def test_normal_2(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        values = [1.1, 2.2, 3.33]
        expected_list = ["1.10", "2.20", "3.33"]

        col_dp.update_header(DataProperty("abc"))
        for value in values:
            col_dp.update_body(DataProperty(value))

        assert len(values) == len(expected_list)
        for value, expected in zip(values, expected_list):
            assert col_dp.dp_to_str(DataProperty(value)) == expected

    def test_normal_dict(self):
        col_dp = ColumnDataProperty(0, float_type=Decimal)
        values = ["abc", {"1": 1}]
        expected_list = ["abc", "{'1': 1}"]

        col_dp.update_header(DataProperty("dict item"))
        for value in values:
            col_dp.update_body(DataProperty(value))

        assert len(values) == len(expected_list)
        for value, expected in zip(values, expected_list):
            assert col_dp.dp_to_str(DataProperty(value)) == expected

    @pytest.mark.parametrize(
        ["values", "expected_list"],
        [
            [[1234, 223, 1234567], ["1,234", "223", "1,234,567"]],
            [[1234.1, 223.33, 1234567.33], ["1,234.10", "223.33", "1,234,567.33"]],
        ],
    )
    def test_normal_format(self, values, expected_list):
        col_dp = ColumnDataProperty(0, format_flags=Format.THOUSAND_SEPARATOR, float_type=Decimal)

        col_dp.update_header(DataProperty("format test"))
        for value in values:
            col_dp.update_body(DataProperty(value))

        assert len(values) == len(expected_list)
        for value, expected in zip(values, expected_list):
            assert col_dp.dp_to_str(DataProperty(value)) == expected



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import abc
from collections.abc import Sequence
from decimal import Decimal
from typing import Any, Final, Optional, Union

from typepy import RealNumber


T = Union[int, float, Decimal]
NAN: Final = Decimal("nan")


class AbstractContainer(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def min_value(self) -> Optional[Decimal]:  # pragma: no cover
        pass

    @property
    @abc.abstractmethod
    def max_value(self) -> Optional[Decimal]:  # pragma: no cover
        pass

    @abc.abstractmethod
    def mean(self) -> Decimal:  # pragma: no cover
        pass

    @abc.abstractmethod
    def update(self, value: Optional[T]) -> None:  # pragma: no cover
        pass

    @abc.abstractmethod
    def merge(self, value: "AbstractContainer") -> None:  # pragma: no cover
        pass

    def __repr__(self) -> str:
        if not self.has_value():
            return "None"

        return ", ".join([f"min={self.min_value}", f"max={self.max_value}"])

    def has_value(self) -> bool:
        return self.min_value is not None and self.max_value is not None

    def is_same_value(self) -> bool:
        return self.has_value() and self.min_value == self.max_value

    def is_zero(self) -> bool:
        return self.has_value() and self.min_value == 0 and self.max_value == 0


class ListContainer(AbstractContainer):
    __slots__ = ("__value_list",)

    @property
    def min_value(self) -> Optional[Decimal]:
        try:
            return min(self.__value_list)
        except ValueError:
            return None

    @property
    def max_value(self) -> Optional[Decimal]:
        try:
            return max(self.__value_list)
        except ValueError:
            return None

    @property
    def value_list(self) -> list[Decimal]:
        return self.__value_list

    def __init__(self, value_list: Optional[list[Decimal]] = None) -> None:
        if value_list is None:
            self.__value_list: list[Decimal] = []
            return

        for value in value_list:
            self.update(value)

    def mean(self) -> Decimal:
        try:
            return Decimal(sum(self.__value_list) / len(self.__value_list))
        except ZeroDivisionError:
            return NAN

    def update(self, value: Union[int, float, Decimal, None]) -> None:
        if value is None:
            return

        store_value = RealNumber(value).try_convert()
        if store_value is None:
            return

        self.__value_list.append(store_value)

    def merge(self, value: "AbstractContainer") -> None:
        if not isinstance(value, ListContainer):
            return

        for v in value.value_list:
            self.update(v)


class MinMaxContainer(AbstractContainer):
    __slots__ = ("__min_value", "__max_value")

    def __init__(self, value_list: Optional[Sequence[Decimal]] = None) -> None:
        self.__min_value: Optional[Decimal] = None
        self.__max_value: Optional[Decimal] = None

        if value_list is None:
            return

        for value in value_list:
            self.update(value)

    @property
    def min_value(self) -> Optional[Decimal]:
        return self.__min_value

    @property
    def max_value(self) -> Optional[Decimal]:
        return self.__max_value

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MinMaxContainer):
            return False

        return all([self.min_value == other.min_value, self.max_value == other.max_value])

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, MinMaxContainer):
            return True

        return any([self.min_value != other.min_value, self.max_value != other.max_value])

    def __contains__(self, x: T) -> bool:
        if self.min_value is None:
            return False

        if self.max_value is None:
            return False

        return self.min_value <= x <= self.max_value

    def diff(self) -> Decimal:
        if self.min_value is None:
            return NAN

        if self.max_value is None:
            return NAN

        try:
            return self.max_value - self.min_value
        except TypeError:
            return NAN

    def mean(self) -> Decimal:
        if self.min_value is None:
            return NAN

        if self.max_value is None:
            return NAN

        try:
            return (self.max_value + self.min_value) * Decimal("0.5")
        except TypeError:
            return NAN

    def update(self, value: Optional[T]) -> None:
        if value is None:
            return

        decimal_value = Decimal(value)

        if self.__min_value is None:
            self.__min_value = decimal_value
        else:
            self.__min_value = min(self.__min_value, decimal_value)

        if self.__max_value is None:
            self.__max_value = decimal_value
        else:
            self.__max_value = max(self.__max_value, decimal_value)

    def merge(self, value: "AbstractContainer") -> None:
        if not isinstance(value, MinMaxContainer):
            return

        self.update(value.min_value)
        self.update(value.max_value)



import pytest
from typepy import Typecode

from dataproperty import Format
from dataproperty._common import DefaultValue
from dataproperty._formatter import Formatter


dt_format = DefaultValue.DATETIME_FORMAT


class TestFormatter_make_format_str:
    @pytest.mark.parametrize(
        ["format_flags", "datetime_format_str", "decimal_places", "typecode", "expected"],
        [
            [None, dt_format, None, Typecode.STRING, "{:s}"],
            [Format.THOUSAND_SEPARATOR, dt_format, None, Typecode.STRING, "{:s}"],
            [None, dt_format, None, Typecode.INTEGER, "{:d}"],
            [Format.THOUSAND_SEPARATOR, dt_format, None, Typecode.INTEGER, "{:,d}"],
            [None, dt_format, 2, Typecode.INTEGER, "{:d}"],
            [None, dt_format, None, Typecode.REAL_NUMBER, "{:f}"],
            [Format.THOUSAND_SEPARATOR, dt_format, None, Typecode.REAL_NUMBER, "{:,f}"],
            [Format.THOUSAND_SEPARATOR, dt_format, 2, Typecode.REAL_NUMBER, "{:,.2f}"],
            [None, dt_format, 2, Typecode.REAL_NUMBER, "{:.2f}"],
            [None, dt_format, None, Typecode.INFINITY, "{:f}"],
            [None, dt_format, None, Typecode.NAN, "{:f}"],
            [None, dt_format, None, Typecode.DATETIME, "{:%Y-%m-%dT%H:%M:%S%z}"],
            [None, "%Y-%m-%d", None, Typecode.DATETIME, "{:%Y-%m-%d}"],
            [None, None, None, Typecode.NONE, "{}"],
            [None, None, None, Typecode.IP_ADDRESS, "{}"],
            [None, None, None, Typecode.BOOL, "{}"],
            [None, None, None, Typecode.DICTIONARY, "{}"],
            [None, None, None, Typecode.LIST, "{}"],
        ],
    )
    def test_normal(self, format_flags, datetime_format_str, decimal_places, typecode, expected):
        formatter = Formatter(format_flags=format_flags, datetime_format_str=datetime_format_str)

        assert formatter.make_format_str(typecode, decimal_places) == expected


class TestFormatter_make_format_map:
    @pytest.mark.parametrize(
        ["format_flags", "datetime_format_str", "decimal_places", "expected"],
        [[None, dt_format, "", {}]],
    )
    def test_normal(self, format_flags, datetime_format_str, decimal_places, expected):
        formatter = Formatter(format_flags=format_flags, datetime_format_str=datetime_format_str)

        assert formatter.make_format_map(decimal_places) == {
            Typecode.INTEGER: "{:d}",
            Typecode.REAL_NUMBER: "{:f}",
            Typecode.INFINITY: "{:f}",
            Typecode.NAN: "{:f}",
            Typecode.DATETIME: "{:%Y-%m-%dT%H:%M:%S%z}",
            Typecode.NONE: "{}",
            Typecode.IP_ADDRESS: "{}",
            Typecode.BOOL: "{}",
            Typecode.DICTIONARY: "{}",
            Typecode.LIST: "{}",
        }



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import pytest

from dataproperty import LineBreakHandling, Preprocessor


class Test_Preprocessor_update:
    def test_normal(self):
        preprocessor = Preprocessor()
        assert preprocessor.strip_str is None
        assert preprocessor.replace_tabs_with_spaces is True
        assert preprocessor.tab_length == 2
        assert preprocessor.line_break_handling is LineBreakHandling.NOP
        assert preprocessor.line_break_repl == " "
        assert preprocessor.dequote is False
        assert preprocessor.is_escape_html_tag is False
        assert preprocessor.is_escape_formula_injection is False

        assert preprocessor.update(
            strip_str='"',
            replace_tabs_with_spaces=False,
            tab_length=4,
            line_break_handling=LineBreakHandling.REPLACE,
            line_break_repl="<br>",
            dequote=True,
            is_escape_html_tag=True,
            is_escape_formula_injection=True,
        )
        assert preprocessor.strip_str == '"'
        assert preprocessor.replace_tabs_with_spaces is False
        assert preprocessor.tab_length == 4
        assert preprocessor.line_break_handling is LineBreakHandling.REPLACE
        assert preprocessor.line_break_repl == "<br>"
        assert preprocessor.dequote is True
        assert preprocessor.is_escape_html_tag is True
        assert preprocessor.is_escape_formula_injection is True

        assert not preprocessor.update(strip_str='"')
        assert preprocessor.update(strip_str="")


class Test_Preprocessor_preprocess:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            ['abc "efg"', 'abc "efg"'],
            ['"abc efg"', "abc efg"],
            ["'abc efg'", "abc efg"],
            ['"abc" "efg"', '"abc" "efg"'],
            ["'abc' 'efg'", "'abc' 'efg'"],
            ["\"abc 'efg'\"", "abc 'efg'"],
        ],
    )
    def test_normal_dequote(self, value, expected):
        preprocessor = Preprocessor(
            dequote=True,
        )
        data, no_ansi_escape_data = preprocessor.preprocess(value)
        assert data == expected


class Test_Preprocessor_preprocess_string:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [{"1": 1}, {"1": 1}],
            [{"1"}, {"1"}],
        ],
    )
    def test_not_str(self, value, expected):
        preprocessor = Preprocessor(dequote=True)
        data, _ = preprocessor._Preprocessor__preprocess_string(value)
        assert data == expected



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from .__version__ import __author__, __copyright__, __email__, __license__, __version__
from ._align import Align
from ._align_getter import align_getter
from ._column import ColumnDataProperty
from ._common import MAX_STRICT_LEVEL_MAP, MIN_STRICT_LEVEL_MAP, NOT_QUOTING_FLAGS, DefaultValue
from ._container import MinMaxContainer
from ._dataproperty import DataProperty
from ._extractor import DataPropertyExtractor, DataPropertyMatrix, MatrixFormatting
from ._formatter import Format
from ._function import calc_ascii_char_width, get_integer_digit, get_number_of_digit
from ._line_break import LineBreakHandling
from ._preprocessor import Preprocessor
from .logger import set_logger


__all__ = (
    "Align",
    "align_getter",
    "ColumnDataProperty",
    "DataProperty",
    "DataPropertyExtractor",
    "DataPropertyMatrix",
    "Format",
    "LineBreakHandling",
    "MatrixFormatting",
    "MinMaxContainer",
    "Preprocessor",
    "calc_ascii_char_width",
    "get_integer_digit",
    "get_number_of_digit",
    "MAX_STRICT_LEVEL_MAP",
    "MIN_STRICT_LEVEL_MAP",
    "NOT_QUOTING_FLAGS",
    "DefaultValue",
    "set_logger",
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__version__",
)



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import pytest

from dataproperty.typing import (
    Bool,
    DateTime,
    Dictionary,
    Infinity,
    Integer,
    IpAddress,
    List,
    Nan,
    NoneType,
    NullString,
    RealNumber,
    String,
    normalize_type_hint,
)


class Test_normalize_type_hint:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            ["bool", Bool],
            ["datetime", DateTime],
            ["dict", Dictionary],
            ["dictionary", Dictionary],
            ["inf", Infinity],
            ["infinity", Infinity],
            ["int", Integer],
            ["int ", Integer],
            ["int_", Integer],
            ["integer", Integer],
            ["ip", IpAddress],
            ["ipaddr", IpAddress],
            ["ipaddress", IpAddress],
            ["list", List],
            ["nan", Nan],
            ["none", NoneType],
            ["nullstr", NullString],
            ["nullstring", NullString],
            ["float", RealNumber],
            ["realnumber", RealNumber],
            ["str", String],
            ["string", String],
            ["", None],
            [None, None],
        ],
    )
    def test_normal(self, value, expected):
        assert normalize_type_hint(value) == expected



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import datetime
import ipaddress
import sys
from decimal import Decimal

import pytest
from tcolorpy import tcolor
from typepy import (
    Bool,
    DateTime,
    Dictionary,
    Integer,
    Nan,
    RealNumber,
    StrictLevel,
    String,
    Typecode,
)

from dataproperty import (
    MIN_STRICT_LEVEL_MAP,
    Align,
    DataProperty,
    DefaultValue,
    Format,
    LineBreakHandling,
    Preprocessor,
)

from .common import get_strict_level_map


dateutil = pytest.importorskip("dateutil", minversion="2.7")

DATATIME_DATA = datetime.datetime(2017, 1, 2, 3, 4, 5)

nan = float("nan")
inf = float("inf")


class Test_DataPeroperty_eq:
    @pytest.mark.parametrize(
        ["lhs", "rhs", "expected"],
        [
            [1, 1, True],
            [1, 2, False],
            [1, 0.1, False],
            [1, True, False],
            [1.1, 1.1, True],
            [1, nan, False],
            [nan, nan, True],
            [None, None, True],
        ],
    )
    def test_normal(self, lhs, rhs, expected):
        lhs = DataProperty(lhs)
        rhs = DataProperty(rhs)

        assert (lhs == rhs) == expected
        assert (lhs != rhs) == (not expected)


class Test_DataPeroperty_data_typecode:
    @pytest.mark.parametrize(
        ["value", "expected_data", "expected_typecode"],
        [
            ["-0.00284241876820074", Decimal("-0.00284241876820074"), Typecode.REAL_NUMBER],
        ],
    )
    def test_normal(self, value, expected_data, expected_typecode):
        dp = DataProperty(value)

        assert dp == dp
        assert dp.data == expected_data
        assert dp.typecode == expected_typecode

    @pytest.mark.parametrize(
        ["value", "is_convert", "expected_data", "expected_typecode"],
        [
            [1.0, True, 1, Typecode.INTEGER],
            [sys.maxsize, True, sys.maxsize, Typecode.INTEGER],
            [-sys.maxsize, False, -sys.maxsize, Typecode.INTEGER],
            [str(-sys.maxsize), True, -sys.maxsize, Typecode.INTEGER],
            [str(sys.maxsize), False, str(sys.maxsize), Typecode.STRING],
            [1.1, True, 1, Typecode.INTEGER],
            [-1.1, False, Decimal("-1.1"), Typecode.REAL_NUMBER],
            [Decimal("1.1"), False, Decimal("1.1"), Typecode.REAL_NUMBER],
            ["1.1", True, 1, Typecode.INTEGER],
            ["-1.1", False, "-1.1", Typecode.STRING],
            ["a", True, "a", Typecode.STRING],
            ["a", False, "a", Typecode.STRING],
            ["", True, "", Typecode.NULL_STRING],
            ["", False, "", Typecode.NULL_STRING],
            [" ", True, " ", Typecode.NULL_STRING],
            [" ", False, " ", Typecode.NULL_STRING],
            ["3.3.5", True, "3.3.5", Typecode.STRING],
            ["51.0.2704.106", True, "51.0.2704.106", Typecode.STRING],
            [True, True, 1, Typecode.INTEGER],
            [False, False, False, Typecode.BOOL],
            ["100-0002", False, "100-0002", Typecode.STRING],
            ["127.0.0.1", True, ipaddress.IPv4Address("127.0.0.1"), Typecode.IP_ADDRESS],
            ["127.0.0.1", False, "127.0.0.1", Typecode.STRING],
            ["::1", True, ipaddress.IPv6Address("::1"), Typecode.IP_ADDRESS],
            ["::1", False, "::1", Typecode.STRING],
            [[], True, [], Typecode.LIST],
            [[], False, [], Typecode.LIST],
            [{}, True, {}, Typecode.DICTIONARY],
            [{}, False, {}, Typecode.DICTIONARY],
            [
                "2017-01-02 03:04:05",
                True,
                datetime.datetime(2017, 1, 2, 3, 4, 5),
                Typecode.DATETIME,
            ],
            [DATATIME_DATA, True, DATATIME_DATA, Typecode.DATETIME],
            ["2017-01-02 03:04:05", False, "2017-01-02 03:04:05", Typecode.STRING],
            [None, True, None, Typecode.NONE],
            [None, False, None, Typecode.NONE],
            ["None", True, "None", Typecode.STRING],
            ["None", False, "None", Typecode.STRING],
            [inf, True, inf, Typecode.INFINITY],
            [inf, False, Decimal(inf), Typecode.INFINITY],
            ["inf", True, Decimal(inf), Typecode.INFINITY],
            ["inf", False, "inf", Typecode.STRING],
            ["nan", False, "nan", Typecode.STRING],
            [
                "Høgskolen i Østfold er et eksempel...",
                True,
                "Høgskolen i Østfold er et eksempel...",
                Typecode.STRING,
            ],
            [
                "Høgskolen i Østfold er et eksempel...",
                False,
                "Høgskolen i Østfold er et eksempel...",
                Typecode.STRING,
            ],
            [
                "新しいテキスト ドキュメント.txt".encode(),
                True,
                "新しいテキスト ドキュメント.txt",
                Typecode.STRING,
            ],
        ],
    )
    def test_normal_strict_map(self, value, is_convert, expected_data, expected_typecode):
        dp = DataProperty(value, strict_level_map=get_strict_level_map(not is_convert))

        assert dp == dp
        assert dp != DataProperty("test for __ne__")
        assert dp.data == expected_data
        assert dp.typecode == expected_typecode

    @pytest.mark.parametrize(
        ["value", "strip_str", "is_strict", "expected_data", "expected_typecode"],
        [
            ['"1"', '"', False, 1, Typecode.INTEGER],
            ['"1"', "", False, '"1"', Typecode.STRING],
            ['"1"', '"', True, "1", Typecode.STRING],
            ['"1"', "", False, '"1"', Typecode.STRING],
            [b"hoge", "a", False, "hoge", Typecode.STRING],
            ["hogea", b"a", False, "hoge", Typecode.STRING],
        ],
    )
    def test_normal_strip_str(self, value, strip_str, is_strict, expected_data, expected_typecode):
        dp = DataProperty(
            value,
            preprocessor=Preprocessor(strip_str=strip_str),
            strict_level_map=get_strict_level_map(is_strict),
        )

        assert dp.data == expected_data
        assert dp.typecode == expected_typecode

    @pytest.mark.parametrize(
        ["value", "type_hint", "strict_level", "expected_typecode"],
        [
            ["2017-01-02 03:04:05", None, StrictLevel.MIN, Typecode.DATETIME],
            ["2017-01-02 03:04:05", None, StrictLevel.MAX, Typecode.STRING],
            ["2017-01-02 03:04:05", DateTime, StrictLevel.MIN, Typecode.DATETIME],
            ["2017-01-02 03:04:05", DateTime, StrictLevel.MAX, Typecode.DATETIME],
            ["2017-01-02 03:04:05", Integer, StrictLevel.MIN, Typecode.DATETIME],
            ["2017-01-02 03:04:05", Integer, StrictLevel.MAX, Typecode.STRING],
            ["1,000,000,000,000", None, StrictLevel.MAX, Typecode.STRING],
            ["1,000,000,000,000", None, StrictLevel.MIN + 1, Typecode.INTEGER],
            ["1,000,000,000,000", None, StrictLevel.MIN, Typecode.INTEGER],
            ["1,000,000,000,000", Integer, StrictLevel.MAX, Typecode.INTEGER],
            ["1,000,000,000,000", Integer, StrictLevel.MIN, Typecode.INTEGER],
            [DATATIME_DATA, None, StrictLevel.MIN, Typecode.DATETIME],
            [DATATIME_DATA, None, StrictLevel.MAX, Typecode.DATETIME],
            [DATATIME_DATA, String, StrictLevel.MIN, Typecode.STRING],
            [DATATIME_DATA, String, StrictLevel.MAX, Typecode.STRING],
            ["100-0002", None, StrictLevel.MIN, Typecode.STRING],
            ["45e76582", None, StrictLevel.MIN, Typecode.INTEGER],
            ["45e76582", None, StrictLevel.MAX, Typecode.STRING],
            ["4.5e-4", None, StrictLevel.MIN, Typecode.INTEGER],
            ["4.5e-4", None, StrictLevel.MIN + 1, Typecode.REAL_NUMBER],
            ["4.5e-4", None, StrictLevel.MAX, Typecode.STRING],
            [1, String, StrictLevel.MAX, Typecode.STRING],
            [1, String, StrictLevel.MIN, Typecode.STRING],
            [float("inf"), RealNumber, StrictLevel.MAX, Typecode.INFINITY],
            [float("inf"), RealNumber, StrictLevel.MIN, Typecode.INFINITY],
            [1, RealNumber, StrictLevel.MAX, Typecode.INTEGER],
            [1, RealNumber, StrictLevel.MIN, Typecode.INTEGER],
            [1.1, Integer, StrictLevel.MAX, Typecode.INTEGER],
            [1.1, Integer, StrictLevel.MIN, Typecode.INTEGER],
            ["true", None, StrictLevel.MAX, Typecode.STRING],
            ["false", None, StrictLevel.MAX, Typecode.STRING],
            ["true", None, StrictLevel.MIN, Typecode.BOOL],
            ["false", None, StrictLevel.MIN, Typecode.BOOL],
            ["true", Bool, StrictLevel.MIN, Typecode.BOOL],
            ["false", Bool, StrictLevel.MIN, Typecode.BOOL],
            [b"hoge", None, StrictLevel.MAX, Typecode.STRING],
            [b"hoge", None, StrictLevel.MIN, Typecode.STRING],
            ['{"foo": 10}', None, StrictLevel.MAX, Typecode.STRING],
            ['{"foo": 10}', None, StrictLevel.MIN, Typecode.DICTIONARY],
            ['{"foo": 10}', Dictionary, StrictLevel.MAX, Typecode.DICTIONARY],
            ['{"foo": 10}', Dictionary, StrictLevel.MIN, Typecode.DICTIONARY],
            [{"foo": 10}, None, StrictLevel.MAX, Typecode.DICTIONARY],
            [{"foo": 10}, None, StrictLevel.MIN, Typecode.LIST],
        ],
    )
    def test_normal_type_hint(self, value, type_hint, strict_level, expected_typecode):
        dp = DataProperty(value, type_hint=type_hint, strict_level_map={"default": strict_level})

        assert dp.typecode == expected_typecode

    @pytest.mark.parametrize(
        ["value", "is_convert", "expected_data", "expected_typecode"],
        [
            [nan, True, nan, Typecode.NAN],
            [nan, False, nan, Typecode.NAN],
            ["nan", True, nan, Typecode.NAN],
        ],
    )
    def test_normal_nan(self, value, is_convert, expected_data, expected_typecode):
        dp = DataProperty(value, strict_level_map=get_strict_level_map(not is_convert))

        assert Nan(dp.data).is_type()
        assert dp.typecode == expected_typecode


class Test_DataPeroperty_to_str:
    @pytest.mark.parametrize(
        ["value", "type_hint", "is_strict", "expected_data", "expected_str"],
        [
            [float("inf"), None, True, Decimal("inf"), "Infinity"],
            [float("inf"), None, False, Decimal("inf"), "Infinity"],
            [float("inf"), RealNumber, True, Decimal("inf"), "Infinity"],
            [float("inf"), RealNumber, False, Decimal("inf"), "Infinity"],
            [float("inf"), String, False, "inf", "inf"],
        ],
    )
    def test_normal(self, value, type_hint, is_strict, expected_data, expected_str):
        dp = DataProperty(
            value, type_hint=type_hint, strict_level_map=get_strict_level_map(is_strict)
        )

        assert dp.data == expected_data
        assert dp.to_str() == expected_str

    @pytest.mark.parametrize(
        ["value", "format_flags", "expected"], [[1234567, Format.THOUSAND_SEPARATOR, "1,234,567"]]
    )
    def test_normal_format_str(self, value, format_flags, expected):
        dp = DataProperty(value, format_flags=format_flags)

        assert dp.to_str() == expected


class Test_DataPeroperty_set_data:
    @pytest.mark.parametrize(
        ["value", "is_convert", "replace_tabs_with_spaces", "tab_length", "expected"],
        [
            ["a\tb", True, True, 2, "a  b"],
            ["\ta\t\tb\tc\t", True, True, 2, "  a    b  c  "],
            ["a\tb", True, True, 4, "a    b"],
            ["a\tb", True, False, 4, "a\tb"],
            ["a\tb", True, True, None, "a\tb"],
        ],
    )
    def test_normal_tab(self, value, is_convert, replace_tabs_with_spaces, tab_length, expected):
        preprocessor = Preprocessor(
            replace_tabs_with_spaces=replace_tabs_with_spaces,
            tab_length=tab_length,
        )
        dp = DataProperty(
            value, preprocessor=preprocessor, strict_level_map=get_strict_level_map(not is_convert)
        )

        assert dp.data == expected


class Test_DataPeroperty_is_escape_html_tag:
    @pytest.mark.parametrize(
        ["value", "is_escape_html_tag", "expected"],
        [
            [
                "<a href='https://google.com'>test</a>",
                True,
                "&lt;a href=&#x27;https://google.com&#x27;&gt;test&lt;/a&gt;",
            ],
            [
                "<a href='https://google.com'>test</a>",
                False,
                "<a href='https://google.com'>test</a>",
            ],
            [True, True, True],
        ],
    )
    def test_normal_tab(self, value, is_escape_html_tag, expected):
        dp = DataProperty(value, preprocessor=Preprocessor(is_escape_html_tag=is_escape_html_tag))

        assert dp.data == expected


class Test_DataPeroperty_float_type:
    @pytest.mark.parametrize(
        ["value", "float_type", "expected"], [[1.1, float, 1.1], [1.1, Decimal, Decimal("1.1")]]
    )
    def test_normal_tab(self, value, float_type, expected):
        dp = DataProperty(value, float_type=float_type)

        assert isinstance(dp.data, float_type)
        assert dp.data == expected


class Test_DataPeroperty_align:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [1, Align.RIGHT],
            [1.1, Align.RIGHT],
            ["a", Align.LEFT],
            [True, Align.LEFT],
            [DATATIME_DATA, Align.LEFT],
            [None, Align.LEFT],
            [inf, Align.LEFT],
            [nan, Align.LEFT],
        ],
    )
    def test_normal(self, value, expected):
        dp = DataProperty(value)
        assert dp.align == expected


class Test_DataPeroperty_len:
    @pytest.mark.parametrize(
        ["value", "expected_acw", "expected_len"],
        [
            [1, 1, None],
            [-1, 2, None],
            [1.0, 1, None],
            [-1.0, 2, None],
            [1.1, 3, None],
            [-1.1, 4, None],
            [12.34, 5, None],
            ["000", 1, None],
            ["123456789", 9, None],
            ["-123456789", 10, None],
            ["45e76582", 8, 8],
            ["a", 1, 1],
            ["a" * 1000, 1000, 1000],
            ["あ", 2, 1],
            [True, 4, None],
            [None, 4, None],
            [inf, 8, None],
            [nan, 3, None],
        ],
    )
    def test_normal(self, value, expected_acw, expected_len):
        dp = DataProperty(value)

        assert dp.ascii_char_width == expected_acw
        assert dp.length == expected_len

    @pytest.mark.parametrize(
        ["value", "expected_acw", "expected_len"],
        [
            [tcolor("0", color="red"), 1, 10],
            [tcolor("12.34", color="red"), 5, 14],
            [tcolor("abc", color="green"), 3, 12],
        ],
    )
    def test_normal_ascii_escape_sequence(self, value, expected_acw, expected_len):
        dp = DataProperty(value)

        assert dp.ascii_char_width == expected_acw
        assert dp.length == expected_len

    @pytest.mark.parametrize(
        ["value", "eaaw", "expected_acw", "expected_len"], [["øø", 1, 2, 2], ["øø", 2, 4, 2]]
    )
    def test_normal_eaaw(self, value, eaaw, expected_acw, expected_len):
        dp = DataProperty(value, east_asian_ambiguous_width=eaaw)

        assert dp.ascii_char_width == expected_acw
        assert dp.length == expected_len

    @pytest.mark.parametrize(["value", "expected"], [[nan, nan]])
    def test_abnormal(self, value, expected):
        Nan(DataProperty(value).length).is_type()

    @pytest.mark.parametrize(
        ["value", "eaaw", "expected"],
        [["øø", None, ValueError], ["øø", 0, ValueError], ["øø", 3, ValueError]],
    )
    def test_exception_eaaw(self, value, eaaw, expected):
        with pytest.raises(expected):
            DataProperty(value, east_asian_ambiguous_width=eaaw).ascii_char_width


class Test_DataPeroperty_is_include_ansi_escape:
    @pytest.mark.parametrize(
        ["value", "expected_acw"],
        [
            [0, False],
            [tcolor("0", color="red"), True],
            [12.34, False],
            [tcolor("12.34", color="red"), True],
            ["abc", False],
            [tcolor("abc", color="green"), True],
        ],
    )
    def test_normal(self, value, expected_acw):
        assert DataProperty(value).is_include_ansi_escape == expected_acw


class Test_DataPeroperty_line_break_handling:
    @pytest.mark.parametrize(
        ["value", "line_break_handling", "expected"],
        [
            ["a\nb", LineBreakHandling.NOP, "a\nb"],
            ["a\nb", LineBreakHandling.REPLACE, "a b"],
            ["a\nb", LineBreakHandling.ESCAPE, "a\\nb"],
            ["a\r\nb", LineBreakHandling.ESCAPE, "a\\r\\nb"],
            [123, LineBreakHandling.ESCAPE, 123],
        ],
    )
    def test_normal(self, value, line_break_handling, expected):
        preprocessor = Preprocessor(line_break_handling=line_break_handling)
        assert DataProperty(value, preprocessor=preprocessor).data == expected


class Test_DataPeroperty_line_break_repl:
    @pytest.mark.parametrize(
        ["value", "line_break_handling", "line_break_repl", "expected"],
        [
            ["a\nb", LineBreakHandling.REPLACE, "<br>", "a<br>b"],
            ["a\n\r\n\nb", LineBreakHandling.REPLACE, "<br>", "a<br><br><br>b"],
            ["a\nb", LineBreakHandling.NOP, "<br>", "a\nb"],
        ],
    )
    def test_normal(self, value, line_break_handling, line_break_repl, expected):
        preprocessor = Preprocessor(
            line_break_handling=line_break_handling, line_break_repl=line_break_repl
        )
        assert DataProperty(value, preprocessor=preprocessor).data == expected


class Test_DataPeroperty_escape_formula_injection:
    @pytest.mark.parametrize(
        ["value", "escape_formula_injection", "expected"],
        [
            ["a+b", True, "a+b"],
            ["=a+b", True, "'=a+b"],
            ["=a+b", False, "=a+b"],
            ["-a+b", True, "'-a+b"],
            ["-a+b", False, "-a+b"],
            ["+a+b", True, "'+a+b"],
            ["+a+b", False, "+a+b"],
            ["@a+b", True, "'@a+b"],
            ["@a+b", False, "@a+b"],
        ],
    )
    def test_normal(self, value, escape_formula_injection, expected):
        preprocessor = Preprocessor(is_escape_formula_injection=escape_formula_injection)
        assert DataProperty(value, preprocessor=preprocessor).data == expected

    @pytest.mark.parametrize(
        ["value", "expected"],
        [[0, 0], [None, None]],
    )
    def test_abnormal(self, value, expected):
        preprocessor = Preprocessor(is_escape_formula_injection=True)
        assert DataProperty(value, preprocessor=preprocessor).data == expected


class Test_DataPeroperty_get_padding_len:
    @pytest.mark.parametrize(
        ["value", "ascii_char_width", "expected"],
        [
            [1, 8, 8],
            ["000", 8, 8],
            ["a" * 1000, 8, 8],
            ["あ", 8, 7],
            ["あ" * 100, 8, 0],
            ["いろは", 8, 5],
            [["side", "where"], 20, 20],
            [["い" * 100, "ろは"], 8, 0],
            [["い", "ろは"], 20, 17],
            [{"1", 1}, 20, 20],
        ],
    )
    def test_normal(self, value, ascii_char_width, expected):
        assert DataProperty(value).get_padding_len(ascii_char_width) == expected

    @pytest.mark.parametrize(
        ["value", "ascii_char_width", "ambiguous_width", "expected"],
        [
            ["aøb", 4, 1, 4],
            ["aøb", 4, 2, 3],
        ],
    )
    def test_normal_east_asian_ambiguous_width(
        self, value, ascii_char_width, ambiguous_width, expected
    ):
        dp = DataProperty(value, east_asian_ambiguous_width=ambiguous_width)
        assert dp.get_padding_len(ascii_char_width) == expected


class Test_DataPeroperty_integer_digits:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [1, 1],
            [1.0, 1],
            [12.34, 2],
        ],
    )
    def test_normal(self, value, expected):
        dp = DataProperty(value)
        assert dp.integer_digits == expected

    @pytest.mark.parametrize(["value"], [[None], [True], [DATATIME_DATA], ["a"], [inf], [nan]])
    def test_abnormal(self, value):
        dp = DataProperty(value)
        Nan(dp.integer_digits).is_type()


class Test_DataPeroperty_decimal_places:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [1, 0],
            [1.0, 0],
            [1.1, 1],
            [12.34, 2],
        ],
    )
    def test_normal(self, value, expected):
        dp = DataProperty(value)
        assert dp.decimal_places == expected

    @pytest.mark.parametrize(["value"], [[None], [True], [DATATIME_DATA], ["a"], [inf], [nan]])
    def test_abnormal(self, value):
        dp = DataProperty(value)
        Nan(dp.decimal_places).is_type()


class Test_DataPeroperty_additional_format_len:
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [2147483648, 0],
            [0, 0],
            [-1, 1],
            [-0.01, 1],
            ["2147483648", 0],
            ["1", 0],
            ["-1", 1],
            ["-0.01", 1],
            [None, 0],
            [True, 0],
            [DATATIME_DATA, 0],
            ["a", 0],
            [inf, 0],
            [nan, 0],
        ],
    )
    def test_normal(self, value, expected):
        dp = DataProperty(value)
        assert dp.additional_format_len == expected


class Test_DataPeroperty_repr:
    @pytest.mark.parametrize(
        ["value", "strict_level_map", "expected"],
        [
            ["100-0004", MIN_STRICT_LEVEL_MAP, 75],
            [{"a": 1}, DefaultValue.STRICT_LEVEL_MAP, 75],
            ["新しいテキスト ドキュメント.txt", DefaultValue.STRICT_LEVEL_MAP, 80],
        ],
    )
    def test_smoke(self, value, strict_level_map, expected):
        dp = DataProperty(value, strict_level_map=strict_level_map)
        assert len(dp.__repr__()) > expected

    @pytest.mark.parametrize(
        ["value", "strict_level_map", "expected"],
        [
            [
                0,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=0, type=INTEGER, align=right, "
                "ascii_width=1, int_digits=1, decimal_places=0, extra_len=0",
            ],
            [
                tcolor("0", color="red"),
                DefaultValue.STRICT_LEVEL_MAP,
                (
                    ("data={}, type=STRING, align=right, ".format(tcolor("0", color="red")))
                    + "ascii_width=1, length=10, int_digits=1, decimal_places=0, "
                    + "extra_len=0"
                ),
            ],
            [
                -1.0,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=-1, type=INTEGER, align=right, "
                "ascii_width=2, int_digits=1, decimal_places=0, extra_len=1",
            ],
            [
                -1.1,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=-1.1, type=REAL_NUMBER, align=right, "
                "ascii_width=4, int_digits=1, decimal_places=1, extra_len=1",
            ],
            [
                -12.234,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=-12.234, type=REAL_NUMBER, align=right, "
                "ascii_width=7, int_digits=2, decimal_places=3, extra_len=1",
            ],
            [
                0.01,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=0.01, type=REAL_NUMBER, align=right, "
                "ascii_width=4, int_digits=1, decimal_places=2, extra_len=0",
            ],
            [
                "0.000000000000001",
                DefaultValue.STRICT_LEVEL_MAP,
                "data=0.000000000000001, type=REAL_NUMBER, align=right, "
                "ascii_width=17, int_digits=1, decimal_places=15, extra_len=0",
            ],
            [
                "abcdefg",
                DefaultValue.STRICT_LEVEL_MAP,
                "data=abcdefg, type=STRING, align=left, ascii_width=7, length=7, extra_len=0",
            ],
            [
                "いろは",
                DefaultValue.STRICT_LEVEL_MAP,
                "data=いろは, type=STRING, align=left, ascii_width=6, length=3, extra_len=0",
            ],
            [
                None,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=None, type=NONE, align=left, ascii_width=4, extra_len=0",
            ],
            [
                True,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=True, type=BOOL, align=left, ascii_width=4, extra_len=0",
            ],
            [
                DATATIME_DATA,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=2017-01-02 03:04:05, type=DATETIME, align=left, "
                "ascii_width=19, extra_len=0",
            ],
            [
                "2017-01-02 03:04:05",
                DefaultValue.STRICT_LEVEL_MAP,
                "data=2017-01-02 03:04:05, type=STRING, align=left, "
                "ascii_width=19, length=19, extra_len=0",
            ],
            [
                "2017-01-02 03:04:05+0900",
                MIN_STRICT_LEVEL_MAP,
                "data=2017-01-02 03:04:05+09:00, type=DATETIME, align=left, "
                "ascii_width=24, extra_len=0",
            ],
            [
                inf,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=Infinity, type=INFINITY, align=left, ascii_width=8, extra_len=0",
            ],
            [
                nan,
                DefaultValue.STRICT_LEVEL_MAP,
                "data=NaN, type=NAN, align=left, ascii_width=3, extra_len=0",
            ],
            [
                ["side", "where"],
                DefaultValue.STRICT_LEVEL_MAP,
                "data=['side', 'where'], type=LIST, align=left, "
                "ascii_width=17, length=2, extra_len=0",
            ],
            [
                ["い", "ろは"],
                DefaultValue.STRICT_LEVEL_MAP,
                "data=['い', 'ろは'], type=LIST, align=left, ascii_width=14, length=2, extra_len=0",
            ],
        ],
    )
    def test_normal(self, value, strict_level_map, expected):
        dp = DataProperty(value, strict_level_map=strict_level_map)

        print(f"[expected] {expected}")
        print(f"[actual]   {dp}")

        assert str(dp) == expected



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import abc
from typing import Optional

from typepy import Typecode

from ._align import Align


class DataPeropertyInterface(metaclass=abc.ABCMeta):
    __slots__ = ()

    @property
    @abc.abstractmethod
    def align(self) -> Align:  # pragma: no cover
        pass

    @property
    @abc.abstractmethod
    def decimal_places(self) -> Optional[int]:  # pragma: no cover
        pass

    @property
    @abc.abstractmethod
    def typecode(self) -> Typecode:  # pragma: no cover
        pass

    @property
    @abc.abstractmethod
    def typename(self) -> str:  # pragma: no cover
        pass



#!/usr/bin/env python3

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import datetime
import sys

from dataproperty import DataPropertyExtractor


def display_dp_matrix_attr(dp_matrix, attr_name):
    """
    show a value associated with an attribute for each
    DataProperty instance in the dp_matrix
    """

    print()
    print(f"---------- {attr_name:s} ----------")
    for dp_list in dp_matrix:
        print([getattr(dp, attr_name) for dp in dp_list])


def main():
    # sample data definitions
    dt = datetime.datetime(2017, 1, 1, 0, 0, 0)
    inf = float("inf")
    nan = float("nan")

    # extract data property for each element from a matrix
    dp_extractor = DataPropertyExtractor()
    dp_matrix = dp_extractor.to_dp_matrix(
        [
            [1, 1.1, "aa", 1, 1, True, inf, nan, dt],
            [2, 2.2, "bbb", 2.2, 2.2, False, "inf", "nan", dt],
            [3, 3.33, "cccc", -3, "ccc", "true", inf, "NAN", "2017-01-01T01:23:45+0900"],
        ]
    )

    print("---------- typename ----------")
    for dp_list in dp_matrix:
        print([dp.typecode.name for dp in dp_list])

    display_dp_matrix_attr(dp_matrix, "data")
    display_dp_matrix_attr(dp_matrix, "align")
    display_dp_matrix_attr(dp_matrix, "ascii_char_width")
    display_dp_matrix_attr(dp_matrix, "integer_digits")
    display_dp_matrix_attr(dp_matrix, "decimal_places")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import itertools

import pytest

from dataproperty import calc_ascii_char_width


nan = float("nan")
inf = float("inf")


class Test_calc_ascii_char_width:
    @pytest.mark.parametrize(
        ["value", "expected"], [["吾輩は猫である", 14], ["いaろbはc", 9], ["abcdef", 6], ["", 0]]
    )
    def test_normal(self, value, expected):
        assert calc_ascii_char_width(value) == expected

    @pytest.mark.parametrize(
        ["value", "ambiguous_width"],
        itertools.product(["Ø", "α", "β", "γ", "θ", "κ", "λ", "π", "ǎ"], [1, 2]),
    )
    def test_normal_east_asian_ambiguous(self, value, ambiguous_width):
        assert calc_ascii_char_width(value, ambiguous_width) == ambiguous_width

    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [b"abcdef", TypeError],
            [None, TypeError],
            [True, TypeError],
            [1, TypeError],
            [nan, TypeError],
        ],
    )
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            calc_ascii_char_width(value)



#!/usr/bin/env python3

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import datetime
import sys

from dataproperty import DataPropertyExtractor


def display_col_dp(dp_list, attr_name):
    """
    show a value associated with an attribute for each
    DataProperty instance in the dp_list
    """

    print()
    print(f"---------- {attr_name:s} ----------")
    print([getattr(dp, attr_name) for dp in dp_list])


def main():
    # sample data definitions
    dt = datetime.datetime(2017, 1, 1, 0, 0, 0)
    inf = float("inf")
    nan = float("nan")
    data_matrix = [
        [1, 1.1, "aa", 1, 1, True, inf, nan, dt],
        [2, 2.2, "bbb", 2.2, 2.2, False, "inf", "nan", dt],
        [3, 3.33, "cccc", -3, "ccc", "true", inf, "NAN", "2017-01-01T01:23:45+0900"],
    ]

    # extract property for each column from a matrix
    dp_extractor = DataPropertyExtractor()
    dp_extractor.headers = ["int", "float", "str", "num", "mix", "bool", "inf", "nan", "time"]
    col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(data_matrix))

    print("---------- typename ----------")
    print([dp.typecode.name for dp in col_dp_list])

    display_col_dp(col_dp_list, "align")
    display_col_dp(col_dp_list, "ascii_char_width")
    display_col_dp(col_dp_list, "decimal_places")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import datetime
from decimal import Decimal

import pytest
from typepy import DateTime, RealNumber, String, Typecode

from dataproperty import (
    Align,
    DataPropertyExtractor,
    Format,
    LineBreakHandling,
    MatrixFormatting,
    Preprocessor,
)

from .common import get_strict_level_map


DATATIME_DATA = datetime.datetime(2017, 1, 2, 3, 4, 5)

nan = float("nan")
inf = float("inf")


@pytest.fixture
def dp_extractor():
    return DataPropertyExtractor()


def datetime_formatter_test(value):
    return value.strftime("%Y%m%d %H%M%S")


def datetime_formatter_tostr_0(value):
    return value.strftime("%Y-%m-%d %H:%M:%S%z")


def datetime_formatter_tostr_1(value):
    return value.strftime("%Y/%m/%d %H:%M:%S")


def trans_func_1(v):
    if v is None:
        return ""
    if v is False:
        return "false"
    if v == 0:
        return 123
    return v


def trans_func_2(v):
    if v == 123:
        return 321
    return v


def nop(v):
    return v


class Test_DataPropertyExtractor_to_dp:
    @pytest.mark.parametrize(
        ["value", "type_value_map", "is_strict", "expected_value", "expected_typecode"],
        [
            [None, {Typecode.NONE: None}, True, None, Typecode.NONE],
            [None, {Typecode.NONE: "null"}, False, "null", Typecode.STRING],
            [None, {Typecode.NONE: ""}, True, "", Typecode.NULL_STRING],
            [None, {Typecode.NONE: 0}, False, 0, Typecode.INTEGER],
            [inf, {Typecode.INFINITY: "INF_1"}, False, "INF_1", Typecode.STRING],
            [inf, {Typecode.INFINITY: "INF_2"}, True, "INF_2", Typecode.STRING],
            [inf, {Typecode.INFINITY: None}, True, None, Typecode.NONE],
            ["inf", {Typecode.INFINITY: "INF_3"}, False, "INF_3", Typecode.STRING],
            ["inf", {Typecode.INFINITY: "INF_4"}, True, "inf", Typecode.STRING],
            ["inf", {Typecode.INFINITY: inf}, False, Decimal("Infinity"), Typecode.INFINITY],
            [nan, {Typecode.NAN: "NAN_1"}, False, "NAN_1", Typecode.STRING],
            [nan, {Typecode.NAN: "NAN_2"}, True, "NAN_2", Typecode.STRING],
            [nan, {Typecode.NAN: None}, True, None, Typecode.NONE],
            ["nan", {Typecode.NAN: "NAN_4"}, False, "NAN_4", Typecode.STRING],
            ["nan", {Typecode.NAN: "NAN_5"}, True, "nan", Typecode.STRING],
        ],
    )
    def test_normal_type_value_map(
        self, dp_extractor, value, type_value_map, is_strict, expected_value, expected_typecode
    ):
        dp_extractor.type_value_map = type_value_map
        dp_extractor.strict_level_map = get_strict_level_map(is_strict)
        dp = dp_extractor.to_dp(value)

        assert dp.data == expected_value
        assert dp.typecode == expected_typecode
        assert isinstance(dp.to_str(), str)

    @pytest.mark.parametrize(
        ["value", "datetime_formatter", "datetime_format_str", "is_strict", "expected"],
        [
            [DATATIME_DATA, datetime_formatter_tostr_0, "s", False, "2017-01-02 03:04:05"],
            ["2017-01-01 00:00:00", datetime_formatter_tostr_1, "s", False, "2017/01/01 00:00:00"],
            [
                "2017-01-01 00:00:00",
                None,
                "%Y-%m-%dT%H:%M:%S",
                False,
                datetime.datetime(2017, 1, 1, 0, 0, 0),
            ],
            ["2017-01-01 00:00:00", None, "s", True, "2017-01-01 00:00:00"],
        ],
    )
    def test_normal_datetime(
        self, dp_extractor, value, datetime_formatter, datetime_format_str, is_strict, expected
    ):
        dp_extractor.datetime_formatter = datetime_formatter
        dp_extractor.datetime_format_str = datetime_format_str
        dp_extractor.strict_level_map = get_strict_level_map(is_strict)
        dp = dp_extractor.to_dp(value)

        assert dp.data == expected

    @pytest.mark.parametrize(
        ["value", "type_hint", "trans_func", "expected"],
        [
            [1, String, nop, "1"],
            [0, String, nop, "0"],
            [None, String, nop, "None"],
            [0, String, trans_func_1, "123"],
            [False, String, trans_func_1, "false"],
            [None, String, trans_func_1, ""],
        ],
    )
    def test_normal_type_hint(self, dp_extractor, value, type_hint, trans_func, expected):
        dp_extractor.register_trans_func(trans_func)
        dp = dp_extractor._DataPropertyExtractor__to_dp(value, type_hint=type_hint)

        assert dp.data == expected

    @pytest.mark.parametrize(
        ["value", "type_hint", "trans_funcs", "expected"],
        [
            [0, String, [trans_func_2, trans_func_1], "321"],
            [0, String, [trans_func_1, trans_func_2], "123"],
        ],
    )
    def test_normal_trans_funcs(self, dp_extractor, value, type_hint, trans_funcs, expected):
        for trans_func in trans_funcs:
            dp_extractor.register_trans_func(trans_func)
        dp = dp_extractor._DataPropertyExtractor__to_dp(value, type_hint=type_hint)

        assert dp.data == expected


class Test_DataPropertyExtractor_to_dp_quoting_flags:
    ALWAYS_QUOTE_FLAG_MAP = {
        Typecode.NONE: True,
        Typecode.INTEGER: True,
        Typecode.REAL_NUMBER: True,
        Typecode.STRING: True,
        Typecode.NULL_STRING: True,
        Typecode.DATETIME: True,
        Typecode.REAL_NUMBER: True,
        Typecode.NAN: True,
        Typecode.BOOL: True,
    }

    @pytest.mark.parametrize(
        ["value", "quoting_flags", "expected"],
        [
            ["string", ALWAYS_QUOTE_FLAG_MAP, '"string"'],
            ['"string"', ALWAYS_QUOTE_FLAG_MAP, '"string"'],
            [' "123"', ALWAYS_QUOTE_FLAG_MAP, ' "123"'],
            ['"string" ', ALWAYS_QUOTE_FLAG_MAP, '"string" '],
            [' "12 345" ', ALWAYS_QUOTE_FLAG_MAP, ' "12 345" '],
        ],
    )
    def test_normal_always_quote(self, dp_extractor, value, quoting_flags, expected):
        dp_extractor.quoting_flags = quoting_flags
        dp = dp_extractor.to_dp(value)

        assert dp.data == expected


class Test_DataPropertyExtractor_to_dp_matrix:
    @pytest.mark.parametrize(
        ["value"],
        [
            [
                [
                    [
                        "山田",
                        "太郎",
                        "2001/1/1",
                        "100-0002",
                        "東京都千代田区皇居外苑",
                        "03-1234-5678",
                    ],
                    [
                        "山田",
                        "次郎",
                        "2001/1/2",
                        "251-0036",
                        "神奈川県藤沢市江の島１丁目",
                        "03-9999-9999",
                    ],
                ]
            ]
        ],
    )
    def test_smoke(self, dp_extractor, value):
        assert len(list(dp_extractor.to_dp_matrix(value))) > 0

    @pytest.mark.parametrize(
        ["value", "type_value_map", "datetime_formatter"],
        [
            [
                [[None, "1"], [1.1, "a"], [nan, inf], ["false", DATATIME_DATA]],
                {Typecode.NONE: "null", Typecode.INFINITY: "INFINITY", Typecode.NAN: "NAN"},
                datetime_formatter_test,
            ]
        ],
    )
    def test_normal(self, dp_extractor, value, type_value_map, datetime_formatter):
        dp_extractor.type_value_map = type_value_map
        dp_extractor.datetime_formatter = datetime_formatter
        dp_matrix = list(dp_extractor.to_dp_matrix(dp_extractor.to_dp_matrix(value)))

        assert len(dp_matrix) == 4

        dp = dp_matrix[0][0]
        assert dp.data == "null"
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

        dp = dp_matrix[0][1]
        assert dp.data == 1
        assert dp.typecode == Typecode.INTEGER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.decimal_places == 0
        assert dp.format_str == "{:d}"

        dp = dp_matrix[1][0]
        assert dp.data == Decimal("1.1")
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.decimal_places == 1
        assert dp.format_str == "{:.1f}"

        dp = dp_matrix[1][1]
        assert dp.data == "a"
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

        dp = dp_matrix[2][0]
        assert dp.data == "NAN"
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

        dp = dp_matrix[2][1]
        assert dp.data == "INFINITY"
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

        dp = dp_matrix[3][0]
        assert dp.data == "false"
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

        dp = dp_matrix[3][1]
        assert dp.data == "20170102 030405"
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

    @pytest.mark.parametrize(["value", "expected"], [[None, []], [[], []], [(), []]])
    def test_empty(self, dp_extractor, value, expected):
        assert dp_extractor.to_dp_matrix(value) == expected


class Test_DataPropertyExtractor_to_dp_list:
    @pytest.mark.parametrize(
        ["value", "float_type"], [[[0.1, Decimal("1.1")], float], [[0.1, Decimal("1.1")], Decimal]]
    )
    def test_normal_float(self, dp_extractor, value, float_type):
        dp_extractor.float_type = float_type
        dp_list = dp_extractor.to_dp_list(value)

        for dp in dp_list:
            assert isinstance(dp.data, float_type)

    @pytest.mark.parametrize(
        ["value", "type_hint", "expected_list"],
        [
            [
                ["2017-01-02 03:04:05", datetime.datetime(2017, 1, 2, 3, 4, 5)],
                None,
                [Typecode.STRING, Typecode.DATETIME],
            ],
            [
                ["2017-01-02 03:04:05", datetime.datetime(2017, 1, 2, 3, 4, 5)],
                DateTime,
                [Typecode.DATETIME, Typecode.DATETIME],
            ],
        ],
    )
    def test_normal_type_hint(self, dp_extractor, value, type_hint, expected_list):
        dp_extractor.default_type_hint = type_hint
        dp_list = dp_extractor.to_dp_list(value)

        for dp, expected in zip(dp_list, expected_list):
            assert dp.typecode == expected

    @pytest.mark.parametrize(
        ["value", "strip_str_header", "strip_str_value", "expected"],
        [
            [['"1"', '"-1.1"', '"abc"'], "", '"', [1, Decimal("-1.1"), "abc"]],
            [['"1"', '"-1.1"', '"abc"'], '"', "", ['"1"', '"-1.1"', '"abc"']],
            [['"1"', '"-1.1"', '"abc"'], None, None, ['"1"', '"-1.1"', '"abc"']],
        ],
    )
    def test_normal_strip_str(
        self, dp_extractor, value, strip_str_header, strip_str_value, expected
    ):
        dp_extractor.strip_str_header = strip_str_header
        dp_extractor.preprocessor = Preprocessor(strip_str=strip_str_value)
        dp_list = dp_extractor.to_dp_list(value)

        for dp, expected_value in zip(dp_list, expected):
            assert dp.data == expected_value

        dp_matrix = dp_extractor.to_dp_matrix([value])
        for dp, expected_value in zip(dp_matrix[0], expected):
            assert dp.data == expected_value

    @pytest.mark.parametrize(
        ["value", "line_break_handling", "expected"],
        [
            [["a\nb", "a\r\nb"], LineBreakHandling.NOP, ["a\nb", "a\r\nb"]],
            [["a\nb", "a\r\nb"], LineBreakHandling.REPLACE, ["a b", "a b"]],
            [["a\nb", "a\r\nb"], LineBreakHandling.ESCAPE, ["a\\nb", "a\\r\\nb"]],
        ],
    )
    def test_normal_line_break_handling(self, dp_extractor, value, line_break_handling, expected):
        dp_extractor.preprocessor = Preprocessor(line_break_handling=line_break_handling)
        dp_list = dp_extractor.to_dp_list(value)

        for dp, value in zip(dp_list, expected):
            assert dp.data == value

    @pytest.mark.parametrize(
        ["value", "line_break_handling", "line_break_repl", "expected"],
        [
            [["a\nb", "a\r\nb"], LineBreakHandling.NOP, "<br>", ["a\nb", "a\r\nb"]],
            [
                ["a\nb", "a\r\nb", "a\r\n\nb"],
                LineBreakHandling.REPLACE,
                "<br>",
                ["a<br>b", "a<br>b", "a<br><br>b"],
            ],
        ],
    )
    def test_normal_line_break_repl(
        self, dp_extractor, value, line_break_handling, line_break_repl, expected
    ):
        dp_extractor.preprocessor = Preprocessor(
            line_break_handling=line_break_handling, line_break_repl=line_break_repl
        )
        dp_list = dp_extractor.to_dp_list(value)

        for dp, value in zip(dp_list, expected):
            assert dp.data == value, value

    @pytest.mark.parametrize(
        ["value", "escape_formula_injection", "expected"],
        [
            [
                ["a+b", "=a+b", "-a+b", "+a+b", "@a+b"],
                True,
                ["a+b", "'=a+b", "'-a+b", "'+a+b", "'@a+b"],
            ],
            [
                ["a+b", "=a+b", "-a+b", "+a+b", "@a+b"],
                False,
                ["a+b", "=a+b", "-a+b", "+a+b", "@a+b"],
            ],
        ],
    )
    def test_normal_escape_formula_injection(
        self, dp_extractor, value, escape_formula_injection, expected
    ):
        dp_extractor.preprocessor = Preprocessor(
            is_escape_formula_injection=escape_formula_injection
        )
        dp_list = dp_extractor.to_dp_list(value)

        for dp, value in zip(dp_list, expected):
            assert dp.data == value, value

    @pytest.mark.parametrize(
        ["value", "expected"],
        [[[0, None], [0, None]]],
    )
    def test_exception_escape_formula_injection(self, dp_extractor, value, expected):
        dp_extractor.preprocessor = Preprocessor(is_escape_formula_injection=True)
        dp_list = dp_extractor.to_dp_list(value)

        for dp, value in zip(dp_list, expected):
            assert dp.data == value, value


class Test_DataPropertyExtractor_to_column_dp_list:
    TEST_DATA_MATRIX = [
        [1, 1.1, "aa", 1, 1, True, inf, nan, datetime.datetime(2017, 1, 1, 0, 0, 0)],
        [2, 2.2, "bbb", 2.2, 2.2, False, "inf", "nan", "2017-01-01T01:23:45+0900"],
        [3, 3.33, "cccc", -3, "ccc", True, "infinity", "NAN", "2017-11-01 01:23:45+0900"],
    ]
    TEST_DATA_MATRIX_TUPLE = (
        (1, 1.1, "aa", 1, 1, True, inf, nan, datetime.datetime(2017, 1, 1, 0, 0, 0)),
        (2, 2.2, "bbb", 2.2, 2.2, False, "inf", "nan", "2017-01-01T01:23:45+0900"),
        (3, 3.33, "cccc", -3, "ccc", True, "infinity", "NAN", "2017-11-01 01:23:45+0900"),
    )

    @pytest.mark.parametrize(
        ["max_workers", "headers", "value"],
        [
            [1, ["i", "f", "s", "if", "mix", "bool", "inf", "nan", "time"], TEST_DATA_MATRIX],
            [4, ["i", "f", "s", "if", "mix", "bool", "inf", "nan", "time"], TEST_DATA_MATRIX],
            [None, None, TEST_DATA_MATRIX],
            [None, [], TEST_DATA_MATRIX],
            [
                None,
                ("i", "f", "s", "if", "mix", "bool", "inf", "nan", "time"),
                TEST_DATA_MATRIX_TUPLE,
            ],
        ],
    )
    def test_normal_default(self, dp_extractor, max_workers, headers, value):
        dp_extractor.max_workers = max_workers
        dp_extractor.headers = headers
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(value))

        assert len(col_dp_list) == 9

        col_idx = 0

        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.INTEGER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.ascii_char_width == 1
        assert dp.decimal_places == 0
        assert dp.format_str == "{:d}"
        assert str(dp) == (
            "column=0, type=INTEGER, align=right, "
            "ascii_width=1, bit_len=2, int_digits=1, decimal_places=0"
        )

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.ascii_char_width == 4
        assert dp.decimal_places == 2
        assert dp.format_str == "{:.2f}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 4
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.ascii_char_width == 4
        assert dp.decimal_places == 1
        assert dp.format_str == "{:.1f}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 3
        assert dp.decimal_places == 1
        assert dp.format_str == "{:s}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.BOOL
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 5
        assert dp.decimal_places is None
        assert dp.format_str == "{}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.INFINITY
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 8
        assert dp.decimal_places is None
        assert dp.format_str == "{:f}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.NAN
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 3
        assert dp.decimal_places is None
        assert dp.format_str == "{:f}"

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 24
        assert dp.decimal_places is None
        assert dp.format_str == "{:s}"

    @pytest.mark.parametrize(
        ["headers", "value"],
        [
            [
                ["i", "f"],
                [
                    [1234, 1234.5],
                    [1234567, 34.5],
                ],
            ],
            [
                [],
                [
                    [1234, 1234.5],
                    [1234567, 34.5],
                ],
            ],
        ],
    )
    def test_normal_format_str(self, dp_extractor, headers, value):
        dp_extractor.format_flags_list = [Format.THOUSAND_SEPARATOR, Format.THOUSAND_SEPARATOR]
        dp_extractor.max_workers = 1
        dp_extractor.headers = headers
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(value))

        assert len(col_dp_list) == 2

        col_idx = 0

        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.INTEGER
        assert dp.format_str == "{:,d}"
        assert dp.ascii_char_width == 9

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.format_str == "{:,.1f}"
        assert dp.ascii_char_width == 7

    @pytest.mark.parametrize(
        ["headers", "value"],
        [
            [["i", "f", "s", "if", "mix", "bool", "inf", "nan", "time"], TEST_DATA_MATRIX],
            [None, TEST_DATA_MATRIX],
            [[], TEST_DATA_MATRIX],
        ],
    )
    def test_normal_not_strict(self, dp_extractor, headers, value):
        dp_extractor.headers = headers
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(value))

        assert len(col_dp_list) == 9

        dp = col_dp_list[0]
        assert dp.typecode == Typecode.INTEGER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.ascii_char_width == 1
        assert dp.decimal_places == 0
        assert dp.format_str == "{:d}"

        dp = col_dp_list[1]
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.align.align_code == Align.RIGHT.align_code
        assert dp.align.align_string == Align.RIGHT.align_string
        assert dp.ascii_char_width == 4
        assert dp.decimal_places == 2
        assert dp.format_str == "{:.2f}"

    def test_normal_column_type_hints(self, dp_extractor):
        data_matrix = [
            [1, "1.1", 1, "2017-01-02 03:04:05"],
            [2, "2.2", 0.1, "2017-01-02 03:04:05"],
        ]
        dp_extractor.headers = ["none", "to_float", "to_str", "to_datetime"]

        dp_extractor.column_type_hints = [None, RealNumber, String, DateTime]
        assert dp_extractor.column_type_hints == [None, RealNumber, String, DateTime]
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(data_matrix))
        assert len(col_dp_list) == 4
        assert col_dp_list[0].typecode == Typecode.INTEGER
        assert col_dp_list[1].typecode == Typecode.REAL_NUMBER
        assert col_dp_list[2].typecode == Typecode.STRING
        assert col_dp_list[3].typecode == Typecode.DATETIME

        dp_extractor.column_type_hints = ["", "float", "str", "datetime"]
        assert dp_extractor.column_type_hints == [None, RealNumber, String, DateTime]
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(data_matrix))
        assert len(col_dp_list) == 4
        assert col_dp_list[0].typecode == Typecode.INTEGER
        assert col_dp_list[1].typecode == Typecode.REAL_NUMBER
        assert col_dp_list[2].typecode == Typecode.STRING
        assert col_dp_list[3].typecode == Typecode.DATETIME

    def test_normal_max_precision(self):
        extractor = DataPropertyExtractor(max_precision=3)
        extractor.headers = ["i", "f"]
        value = [
            [1234, 0.0000000001],
            [1234567, 34.5],
        ]

        col_dp_list = extractor.to_column_dp_list(extractor.to_dp_matrix(value))
        assert len(col_dp_list) == 2

        col_idx = 0
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.INTEGER
        assert dp.decimal_places == 0

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.decimal_places == 3

        # test setter
        extractor.max_precision = 1
        col_dp_list = extractor.to_column_dp_list(extractor.to_dp_matrix(value))
        assert len(col_dp_list) == 2

        col_idx = 0
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.INTEGER
        assert dp.decimal_places == 0

        col_idx += 1
        dp = col_dp_list[col_idx]
        assert dp.column_index == col_idx
        assert dp.typecode == Typecode.REAL_NUMBER
        assert dp.decimal_places == 1

    def test_normal_nan_inf(self, dp_extractor):
        dp_extractor.headers = ["n", "i"]
        col_dp_list = dp_extractor.to_column_dp_list(
            dp_extractor.to_dp_matrix([[nan, inf], ["nan", "inf"]])
        )

        assert len(col_dp_list) == 2

        dp = col_dp_list[0]
        assert dp.typecode == Typecode.NAN
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 3
        assert dp.decimal_places is None

        dp = col_dp_list[1]
        assert dp.typecode == Typecode.INFINITY
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 8
        assert dp.decimal_places is None

    @pytest.mark.parametrize(["ambiguous_width"], [[2], [1]])
    def test_normal_east_asian_ambiguous_width(self, dp_extractor, ambiguous_width):
        dp_extractor.headers = ["ascii", "eaa"]
        dp_extractor.east_asian_ambiguous_width = ambiguous_width
        col_dp_list = dp_extractor.to_column_dp_list(
            dp_extractor.to_dp_matrix([["abcdefg", "Øαββ"], ["abcdefghij", "ØØ"]])
        )

        assert len(col_dp_list) == 2

        dp = col_dp_list[0]
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 10
        assert dp.decimal_places is None

        dp = col_dp_list[1]
        assert dp.typecode == Typecode.STRING
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 4 * ambiguous_width
        assert dp.decimal_places is None

    def test_normal_empty_value(self, dp_extractor):
        dp_extractor.headers = ["a", "22", "cccc"]
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(None))

        dp = col_dp_list[0]
        assert dp.typecode == Typecode.NONE
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 1
        assert dp.decimal_places is None
        assert dp.format_str == "{}"

        dp = col_dp_list[1]
        assert dp.typecode == Typecode.NONE
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 2
        assert dp.decimal_places is None
        assert dp.format_str == "{}"

        dp = col_dp_list[2]
        assert dp.typecode == Typecode.NONE
        assert dp.align.align_code == Align.LEFT.align_code
        assert dp.align.align_string == Align.LEFT.align_string
        assert dp.ascii_char_width == 4
        assert dp.decimal_places is None
        assert dp.format_str == "{}"


class Test_DataPropertyExtractor_matrix_formatting:
    TEST_DATA_MATRIX_NORMAL_COL3 = [["a", 0, "aa"], ["b", 1, "bb"], ["c", 2, "ccc"]]
    TEST_DATA_MATRIX_NOUNIFORM_COL1 = [["a", 0], ["b", 1, "bb"], ["c", 2, "ccc", 0.1], ["d"]]

    @pytest.mark.parametrize(
        ["headers", "value", "matrix_formatting", "expected"],
        [
            [None, TEST_DATA_MATRIX_NOUNIFORM_COL1, MatrixFormatting.TRIM, 1],
            [["a", "b"], TEST_DATA_MATRIX_NORMAL_COL3, MatrixFormatting.TRIM, 2],
            [None, TEST_DATA_MATRIX_NOUNIFORM_COL1, MatrixFormatting.FILL_NONE, 4],
            [["a", "b", "c"], TEST_DATA_MATRIX_NORMAL_COL3, MatrixFormatting.FILL_NONE, 3],
            [["a", "b", "c"], TEST_DATA_MATRIX_NOUNIFORM_COL1, MatrixFormatting.HEADER_ALIGNED, 3],
            [
                ["a", "b", "c", "d", "e"],
                TEST_DATA_MATRIX_NOUNIFORM_COL1,
                MatrixFormatting.HEADER_ALIGNED,
                5,
            ],
        ],
    )
    def test_normal_matrix_formatting(
        self, dp_extractor, headers, value, matrix_formatting, expected
    ):
        dp_extractor.headers = headers
        dp_extractor.matrix_formatting = matrix_formatting
        col_dp_list = dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(value))

        assert len(col_dp_list) == expected

    @pytest.mark.parametrize(
        ["headers", "value", "matrix_formatting", "expected"],
        [
            [
                ["i", "f", "s", "if", "mix"],
                TEST_DATA_MATRIX_NOUNIFORM_COL1,
                MatrixFormatting.EXCEPTION,
                ValueError,
            ]
        ],
    )
    def test_exception_matrix_formatting(
        self, dp_extractor, headers, value, matrix_formatting, expected
    ):
        dp_extractor.headers = headers
        dp_extractor.matrix_formatting = matrix_formatting

        with pytest.raises(expected):
            dp_extractor.to_column_dp_list(dp_extractor.to_dp_matrix(value))


class Test_DataPropertyExtractor_update_preprocessor:
    def test_normal(self, dp_extractor):
        assert dp_extractor.preprocessor.strip_str is None
        assert dp_extractor.preprocessor.replace_tabs_with_spaces is True
        assert dp_extractor.preprocessor.tab_length == 2
        assert dp_extractor.preprocessor.line_break_handling is LineBreakHandling.NOP
        assert dp_extractor.preprocessor.line_break_repl == " "
        assert dp_extractor.preprocessor.is_escape_html_tag is False
        assert dp_extractor.preprocessor.is_escape_formula_injection is False

        dp_extractor.update_preprocessor(
            strip_str='"',
            replace_tabs_with_spaces=False,
            tab_length=4,
            line_break_handling=LineBreakHandling.REPLACE,
            line_break_repl="<br>",
            is_escape_html_tag=True,
            is_escape_formula_injection=True,
        )
        assert dp_extractor.preprocessor.strip_str == '"'
        assert dp_extractor.preprocessor.replace_tabs_with_spaces is False
        assert dp_extractor.preprocessor.tab_length == 4
        assert dp_extractor.preprocessor.line_break_handling is LineBreakHandling.REPLACE
        assert dp_extractor.preprocessor.line_break_repl == "<br>"
        assert dp_extractor.preprocessor.is_escape_html_tag is True
        assert dp_extractor.preprocessor.is_escape_formula_injection is True



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import os.path
from typing import Final

import setuptools


MODULE_NAME: Final = "DataProperty"
REPOSITORY_URL: Final = f"https://github.com/thombashi/{MODULE_NAME:s}"
MISC_DIR: Final = "misc"
REQUIREMENT_DIR: Final = "requirements"

pkg_info: dict[str, str] = {}


def get_release_command_class() -> dict[str, type[setuptools.Command]]:
    try:
        from releasecmd import ReleaseCommand
    except ImportError:
        return {}

    return {"release": ReleaseCommand}


with open(os.path.join(MODULE_NAME.lower(), "__version__.py")) as f:
    exec(f.read(), pkg_info)

with open("README.rst", encoding="utf8") as f:
    long_description = f.read()

with open(os.path.join(MISC_DIR, "summary.txt"), encoding="utf8") as f:
    summary = f.read().strip()

with open(os.path.join(REQUIREMENT_DIR, "requirements.txt")) as f:
    install_requires = [line.strip() for line in f if line.strip()]

with open(os.path.join(REQUIREMENT_DIR, "test_requirements.txt")) as f:
    tests_requires = [line.strip() for line in f if line.strip()]

setuptools.setup(
    name=MODULE_NAME,
    url=REPOSITORY_URL,
    author=pkg_info["__author__"],
    author_email=pkg_info["__email__"],
    description=summary,
    include_package_data=True,
    keywords=[
        "data",
        "library",
        "property",
    ],
    license=pkg_info["__license__"],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    maintainer=pkg_info["__author__"],
    maintainer_email=pkg_info["__email__"],
    packages=setuptools.find_packages(exclude=["test*"]),
    package_data={MODULE_NAME: ["py.typed"]},
    project_urls={
        "Changelog": f"{REPOSITORY_URL:s}/releases",
        "Source": REPOSITORY_URL,
        "Tracker": f"{REPOSITORY_URL:s}/issues",
    },
    python_requires=">=3.9",
    install_requires=install_requires,
    extras_require={
        "logging": ["loguru>=0.4.1,<1"],
        "test": tests_requires,
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    cmdclass=get_release_command_class(),
    zip_safe=False,
)



"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from ._null_logger import NullLogger


MODULE_NAME = "dataproperty"

try:
    from loguru import logger  # type: ignore

    logger.disable(MODULE_NAME)
except ImportError:
    logger = NullLogger()


def set_logger(is_enable: bool, propagation_depth: int = 1) -> None:
    if is_enable:
        logger.enable(MODULE_NAME)
    else:
        logger.disable(MODULE_NAME)
